from collections import defaultdict #, OrderedDict
from datetime import datetime
from urllib.error import URLError
from urllib.request import Request, urlopen
import csv
import functools
import getopt
import io
import json
import os
import re
import sys


covid_data_file = 'covid19_data.json'

def get_covid_data():
    req = Request('https://github.com/CSSEGISandData/COVID-19/tree/master/csse_covid_19_data/csse_covid_19_daily_reports')
    try:
        with urlopen(req) as response:
            html = response.read().decode('utf-8')
    except URLError as e:
        print(e.reason)
        raise

    links = re.findall(
        'CSSEGISandData/COVID-19/blob/master/csse_covid_19_data/csse_covid_19_daily_reports.*\.csv(?=\")',
        html)

    if os.path.isfile(covid_data_file):
        with open(covid_data_file) as fh:
            old_data = json.load(fh)
    else:
        old_data = {}

    new_data = {}

    for link in links:
        link = f"https://raw.githubusercontent.com/{link.replace('/blob','')}"
        key = link.replace('.csv','').rpartition('/')[-1]

        if key not in old_data:
            print(f'Downloading covid-19 reports: {key}')
            req = Request(link)
            try:
                with urlopen(req) as response:
                    resp = response.read().decode('utf-8-sig')
            except URLError as e:
                print(e.reason)
                raise
            csv_reader = csv.DictReader(io.StringIO(resp), delimiter=',')
            # new_data.update({key: [row for row in csv_reader]})
            for row in csv_reader:
                admin = row.get('Admin2', '')
                province = row.get('Province/State') or row.get('Province_State')
                country = row.get('Country/Region') or row.get('Country_Region')
                confirmed = row.get('Confirmed') or 0
                deaths = row.get('Deaths') or 0
                recovered = row.get('Recovered') or 0
                active = row.get('Active', 0)
                last_update = row.get('Last Update') or row.get('Last_Update')
                if not new_data.get(key):
                    new_data[key]=[{
                        'Admin2': admin,
                        'Province_State': province,
                        'Country_Region': country,
                        'Confirmed': int(confirmed),
                        'Deaths': int(deaths),
                        'Recovered': int(recovered),
                        'Active': int(active),
                        'Last_Update': last_update
                    }]
                else:
                    new_data[key].append({
                        'Admin2': admin,
                        'Province_State': province,
                        'Country_Region': country,
                        'Confirmed': int(confirmed),
                        'Deaths': int(deaths),
                        'Recovered': int(recovered),
                        'Active': int(active),
                        'Last_Update': last_update
                    })

    if new_data:
        old_data.update(new_data)
        with open(covid_data_file, "w", encoding='utf-8') as write_file:
            json.dump(old_data, write_file, indent=4)


def load_data():
    if os.path.isfile(covid_data_file):
        mtime = os.path.getmtime(covid_data_file)
        when = datetime.fromtimestamp(mtime)
        now = datetime.now()
        td = now - when
        if td.seconds > 3600:
            print('Checking for updates... Please wait ☻\n')
            get_covid_data()
            os.utime(covid_data_file)
    else:
        get_covid_data()
    with open(covid_data_file) as fh:
        json_data = json.load(fh)
    return json_data


def get_countries(date=None):
    data = load_data()
    if date:
        try:
            data = data[date]
        except KeyError:
            raise KeyError('There is no data for a given date.')
    else:
        data = data[list(data)[-1]]
    countries = {}
    for country in data:
        if country['Country_Region'] not in countries:
            countries.update({country['Country_Region']: {
                                  'Confirmed': country['Confirmed'],
                                  'Deaths': country['Deaths'],
                                  'Recovered': country['Recovered'],
                                  'Active': country['Active'],
                                  'Last_Update': country['Last_Update']}})
        else:
            countries[country['Country_Region']]['Confirmed']+=country['Confirmed']
            countries[country['Country_Region']]['Deaths']+=country['Deaths']
            countries[country['Country_Region']]['Recovered']+=country['Recovered']
            countries[country['Country_Region']]['Active']+=country['Active']
    # return OrderedDict(sorted(countries.items(), key=lambda i: i[1]['Confirmed'], reverse=True))
    return {k:v for k, v in
            sorted(countries.items(), key=lambda i: i[1]['Confirmed'], reverse=True)}


def get_country(name, date=None):
    data = load_data()
    if date:
        try:
            countries = data[date]
        except KeyError:
            raise KeyError('There is no data for a given date.')
    else:
        countries = data[list(data)[-1]]

    # for country in countries:
    #     if country['Country_Region'] == name:
    #         return [i for i in country.values() if i != '']
    country_data = {}
    for country in countries:
        if country['Country_Region'] == name:
            if country['Country_Region'] not in country_data.values():
                country_data['Country']=country['Country_Region']
                country_data['Confirmed']=country['Confirmed']
                country_data['Deaths']=country['Deaths']
                country_data['Recovered']=country['Recovered']
                country_data['Active']=country['Active']
                country_data['Last_Update']=country['Last_Update']
            else:
                country_data['Confirmed']+=country['Confirmed']
                country_data['Deaths']+=country['Deaths']
                country_data['Recovered']+=country['Recovered']
                country_data['Active']+=country['Active']
    return country_data


def get_top(case='Confirmed', num=10):
    case = case.title()
    data = load_data()
    top = {}
    for country in data[list(data)[-1]]:
        if country['Country_Region'] not in top:
            top[country['Country_Region']]=country[case]
        else:
            top[country['Country_Region']]+=country[case]

    return {k:v for k, v in
            sorted(top.items(), key=lambda x: x[1], reverse=True)[:num]}


@functools.lru_cache(maxsize=32)
def get_new_cases(case='Confirmed'):
    case = case.title()
    data = load_data()
    new = {}
    for country1 in data[list(data)[-1]]:
        for country2 in data[list(data)[-2]]:
            if country1['Admin2'] == country2['Admin2'] and \
               country1['Province_State'] == country2['Province_State'] and \
               country1['Country_Region'] == country2['Country_Region']:
                if country1[case] > country2[case]:
                    if country1['Country_Region'] not in new:
                        new[country1['Country_Region']]=(country1[case] - country2[case])
                    else:
                        new[country1['Country_Region']]+=(country1[case] - country2[case])
    # return OrderedDict(sorted(new.items(), key=lambda i: i[1], reverse=True))
    return {k:v for k, v in sorted(new.items(), key=lambda i: i[1], reverse=True)}


def get_total():
    data = load_data()
    total = defaultdict(lambda:0)
    for country in data[list(data)[-1]]:
        total['Confirmed']+=country['Confirmed']
        total['Deaths']+=country['Deaths']
        total['Recovered']+=country['Recovered']
        total['Active']+=country['Active']
    return total


def help():
    usage = """Usage:

    -d mm-dd-yyyy or --date mm-dd-yyyy
        ▶ print the report for a given date

    -c 'Country Name' or --country 'Country Name'
        ▶ print the report for the specified country

    -c 'Country Name' mm-dd-yyyy
        ▶ print the report for the specified country for a given date

    -t case or --top case
        ▶ print ten countries with the most cases (confirmed, deaths, recovered)

    -n case or --new case
        ▶ print countries with new cases (confirmed, deaths, recovered)
    """
    return usage


if __name__ == '__main__':
    opts, args = getopt.getopt(
            sys.argv[1:],
            'd:c:t:n:sh',
            ['date', 'country=', 'top=', 'new=', 'sum', 'help'])
    if len(opts) == 0:
        print(f"\n \033[1m{'COUNTRY':<35}{'CONFIRMED':<20}{'DEATHS':<20}{'RECOVERED':<20}ACTIVE\033[0m\n")
        for country, cases in get_countries().items():
            print(f" {country:<35}{cases['Confirmed']:<20}{cases['Deaths']:<20}{cases['Recovered']:<20}{cases['Active']:<6}")
    else:
        for opt, arg in opts:
            if opt == '-h' or opt == '--help':
                print(help())
            elif opt == '-d' or opt == '--date':
                print(f"\n \033[1m{'COUNTRY':<35}{'CONFIRMED':<20}{'DEATHS':<20}{'RECOVERED':<20}ACTIVE\033[0m\n")
                for country, cases in get_countries(arg).items():
                    print(f" {country:<35}{cases['Confirmed']:<20}{cases['Deaths']:<20}{cases['Recovered']:<20}{cases['Active']:<6}")
            elif opt == '-c' or opt == '--country':
                if len(args) != 0:
                    for k, v in get_country(arg, args[0]).items():
                        print(f'{k:<20} {v}')
                else:
                    for k, v in get_country(arg).items():
                        print(f'{k:<20} {v}')
            elif opt == '-t' or opt == '--top':
                print(f"{' Country':<30}{arg:<20}New {arg:<20}\n")
                for country, case in get_top(arg).items():
                    print(f" {country:<30}{case:<20}{get_new_cases(arg).get(country, 0)}")
            elif opt == '-n' or opt == '--new':
                for country, case in get_new_cases(arg).items():
                    print(f'{country:<30}{case}')
            elif opt == '-s' or opt == '--sum':
                for k, v in get_total().items():
                    print(f" {k:<20}{v}")
